
# -*- coding: utf-8 -*-

from lambda_parser import parse

# ---------------------------------------------------------
# Alias

aliases = {}

def add_alias(s, defi):
    aliases[s] = parse(defi)

# ---------------------------------------------------------
# Création d'objets

def new_var(x): return [0, x]
def new_app(M, N): return [1, M, N]
def new_abs(x, M): return [2, x, M]


# La fonction prod prend en paramètre une liste [M_1,...,M_n] 
# de termes. Elle renvoie le terme $M_1... M_n$.

def prod(ts):
    n = len(ts)
    if n == 0:
        raise Exception('Empty List of Terms')
    else:
        t = ts[0]
        for k in range(1, n):
            t = new_app(t, ts[k])
        return t

# ---------------------------------------------------------
# Les fonctions ci-dessous permettent de tester le "type" d'un terme.

def is_var(P): return P[0] == 0
def is_app(P): return P[0] == 1
def is_abs(P): return P[0] == 2
def is_alias(P): return P[0] == 3

# ---------------------------------------------------------
# Et les suivantes permettent d'extraire les composantes d'un terme.

def var(P):
    if is_var(P) or is_abs(P) or is_alias(P): return P[1]
    else: 
        raise Exception('Applications have no variable')

def left(P):
    if is_app(P): return P[1]
    elif is_var(P) or is_alias(P): 
        raise Exception('Variables and aliases have no left field')
    else: 
        raise Exception('Abstractions have no left field')

def right(P):
    if is_app(P): return P[2]
    elif is_var(P) or is_alias(P): 
        raise Exception('Variables and aliases have no right field')
    else: 
        raise Exception('Abstractions have no right field')

def expr(P):
    if is_abs(P): return P[2]
    elif is_var(P) or is_alias(P): 
        raise Exception('Variables and aliases have no expr field')
    else: 
        raise Exception('Applications have no expr field')

# ---------------------------------------------------------
# Conversion en chaîne de caractères

def to_string(P):
    if is_var(P): return var(P)
    elif is_alias(P): return var(P) + ' '
    elif is_app(P):
        M = left(P)
        N = right(P)
        s1 = to_string(M)
        s2 = to_string(N)
        if is_abs(M):
             s1 = '(' + s1 + ')'
        if is_app(N)or is_abs(N):
             s2 = '(' + s2 + ')'
        return s1 + s2
    else:
        return u'\u03BB' + var(P) + '.' + to_string(expr(P))

def print_term(t):
    print(to_string(t))

def print_aliases():
    for A in aliases:
        print(A, to_string(aliases[A]))

# ---------------------------------------------------------
# Macro-expansion

# Remplacer les alias de P par leur définition
def expand1(P):
    if is_var(P): return (P, False)
    elif is_alias(P): return (aliases[var(P)], True)
    elif is_app(P):
        M, b1 = expand1(left(P))
        N, b2 = expand1(right(P))
        return (new_app(M, N), b1 or b2)
    else:
        M, b = expand1(expr(P))
        x = var(P)
        return (new_abs(x, M), b)

# Application répétée de expand1 jusqu'à ce qu'il n'y ait
# plus d'alias
def expand(P):
    while True:
        P, b = expand1(P)
        if not b: break
    return P

# ---------------------------------------------------------
# Variables libres, variables liées

def is_free(x, P):
    if is_var(P): return x == var(P)
    elif is_app(P):
        return is_free(x, left(P)) or is_free(x, right(P))
    else:
        return x != var(P) and is_free(x, expr(P))

def is_bound(x, P):
    if is_var(P): return False
    elif is_app(P):
        return is_bound(x, left(P)) or is_bound(x, right(P))
    else:
        return x == var(P) or is_bound(x, expr(P))

def free_vars(P):
    if is_var(P): return set([var(P)])
    elif is_app(P):
        return free_vars(left(P)).union(free_vars(right(P)))
    else:
        return free_vars(expr(P)).difference(set([var(P)]))

def bound_vars(P):
    if is_var(P): return set([])
    elif is_app(P):
        return bound_vars(left(P)).union(bound_vars(right(P)))
    else:
        return bound_vars(expr(P)).union(set([var(P)]))

# ---------------------------------------------------------
# Substitution

class VarProvider:

    def __init__(self, prefix='v'):
        self.prefix = prefix
        self.count = 0

    def reset(self): self.count = 0

    def fresh_var(self):
        self.count += 1
        return self.prefix + str(self.count)

def subs(M, x, N, vp):
    if is_var(M):
        y = var(M)
        if x == y: return N
        else: return M
    elif is_app(M):
        P = left(M)
        Q = right(M)
        return new_app(subs(P, x, N, vp), subs(Q, x, N, vp))
    else:
        y = var(M)
        P = expr(M)
        if x == y: return M
        elif not(is_free(x, P)): return M
        elif not(is_free(y, N)):
            return new_abs(y, subs(P, x, N, vp))
        else:
            z = vp.fresh_var()
            P1 = subs(P, y, new_var(z), vp)
            P2 = subs(P1, x, N, vp)
            return new_abs(z, P2)

# ---------------------------------------------------------
# Beta-réduction
# Stratégie : on réduit le redex le plus à gauche

def beta_reduce(P):
    if is_var(P):
        return (P, False)
    elif is_app(P):
        if is_abs(left(P)):
            x = var(left(P))
            A = expr(left(P))
            vp = VarProvider()
            return (subs(A, x, right(P), vp), True)
        else:
            P1, b = beta_reduce(left(P))
            if b: return (new_app(P1, right(P)), True)
            else:
                Q1, b1 = beta_reduce(right(P))
                return (new_app(left(P), Q1), b1)
    else:
        Q, b = beta_reduce(expr(P))
        return (new_abs(var(P), Q), b)

# ---------------------------------------------------------
# Forme normale

# Au maximum max_red beta-réductions
# Si trace = True, affiche les valeurs des beta-réductions
# intermédiaires.

def normal_form(P, max_red=1000, trace=True):
    b = True
    count = 0
    if trace: print(count, to_string(P))
    while b and count < max_red:
        P, b = beta_reduce(P)
        if b: count += 1
        if trace and b: print(count, to_string(P))
    if trace:
        print(20 * '-')
    print('%d réductions\n' % count)
    return P

# ---------------------------------------------------------
# On regroupe le tout

def do(s, **options):
    return normal_form(expand(parse(s)), **options)

